from yta_validation.parameter import ParameterValidator
from abc import ABC, abstractmethod


class ProgressMapping(ABC):
    """
    *Abstract class*

    Abstract class to represent a way of mapping the
    progress, being able to transform the received
    value into a new one by applying the specific
    formula we define on each implementation.

    Having a progress from 0 to 1, this function will
    define the way we go through it.
    """
    
    @abstractmethod
    def map(
        self,
        u: float,
        **kwargs
    ) -> float:
        """
        Map the normalized progress value `u` (that
        must be in the range `[0.0, 1.0]`) into another
        normalized progress value (also in the range
        `[0.0, 1.0]`).
        """
        ParameterValidator.validate_mandatory_number_between('u', u, 0.0, 1.0)

        return self._map(
            u = u,
            **kwargs
        )

    @abstractmethod
    def _map(
        self,
        u: float,
        **kwargs
    ) -> float:
        """
        The internal function that will actually map the
        `u` value provided to the one that should be used
        according to the specific implementation.
        """
        pass



"""
A module to contain the string names of the
easing functions but as Enums to be able to
choose them easy.
"""
from yta_constants.enum import YTAEnum as Enum


class EasingFunctionName(Enum):
    """
    Enum to represent the string names of the different
    easing functions we have registered in our program.
    """

    LINEAR = 'linear'
    SLOW_INTO = 'slow_into'
    SMOOTH = 'smooth'
    SMOOTH_STEP = 'smooth_step'
    SMOOTHER_STEP = 'smoother_step'
    SMOOTHERER_STEP = 'smootherer_step'


# TODO: Maybe we need to indicate if the value
# got clamped or not
def clamp_value_normalized(
    value_normalized: float
) -> float:
    """
    Clamp the `value_normalized` provided, which means
    transforming it into a value that fits in the
    `[0.0, 1.0]` range.
    """
    return (
        1.0
        if value_normalized > 1.0 else
        0.0
        if value_normalized < 0.0 else
        value_normalized
    )


"""
Module to include the different progress mapping
functions we have.
"""
from yta_math_easings.progress_mapping.linear import LinearProgressMapping
from yta_math_easings.progress_mapping.reverse import ReverseProgressMapping
from yta_math_easings.progress_mapping.pingpong import PingPongProgressMapping


__all__ = [
    'LinearProgressMapping',
    'ReverseProgressMapping',
    'PingPongProgressMapping'
]


from yta_math_easings.animated.abstract import EasingAnimatedFunction
from yta_math_easings.smootherer_step import SmoothererStepEasing


class SmoothererStepAnimatedEasing(EasingAnimatedFunction):
    """
    A smootherer step function but animated, including a
    range of values and a duration.

    The formula:
    ```
    (
        35 * n**4 - 84 * n**5 + 70 * n**6 - 20 * n**7
        if 0 < n < 1 else
        1
        if n >= 1 else
        0
    )
    """
    may_overshoot: bool = False
    _easing_cls: type['EasingFunction'] = SmoothererStepEasing

    def ease(
        self,
        t_normalized: float,
        do_clamp: bool = True
    ) -> float:
        return super().ease(
            t_normalized = t_normalized,
            do_clamp = do_clamp
        )


from yta_math_easings.abstract import EasingFunction
from yta_math_easings.enums import EasingFunctionName


class SmoothStepEasing(EasingFunction):
    """
    A smooth step easing function.

    Implementation of the 1st order SmoothStep sigmoid function.
    The 1st derivative (speed) is zero at the endpoints.
    https://en.wikipedia.org/wiki/Smoothstep

    The formula:
    ```
    (
        0
        if t_normalized <= 0 else
        (
            3 * t_normalized**2 - 2 * t_normalized**3
            if t_normalized < 1 else
            1
        )
    )
    ```
    """
    _name: str = EasingFunctionName.SMOOTH_STEP.value

    def ease(
        self,
        t_normalized: float,
        do_clamp: bool = True
    ) -> float:
        return super().ease(
            t_normalized = t_normalized,
            do_clamp = do_clamp
        )

    def _ease(
        self,
        t_normalized: float
    ) -> float:
        return (
            0
            if t_normalized <= 0 else
            (
                3 * t_normalized**2 - 2 * t_normalized**3
                if t_normalized < 1 else
                1
            )
        )


from yta_math_easings.enums import EasingFunctionName
from yta_math_easings.utils import clamp_value_normalized
from yta_validation.parameter import ParameterValidator
from yta_programming.decorators.requires_dependency import requires_dependency
from abc import ABC, abstractmethod

import numpy as np


class EasingFunction(ABC):
    """
    *Abstract class*

    Abstract class to be inherited by the specific
    easing implementations.

    An easing function that is able to calculate the
    normalized value associated to the `t_normalized`
    provided.

    If you need to use real values or duration, check
    the `EasingAnimatedFunction` class instead.
    """
    _registry: dict[str, type['EasingFunction']] = {}
    """
    *For internal use only*
    
    A register to simplify the way we instantiate
    new classes from strings.
    """
    _name: str
    """
    *For internal use only*

    The string name of the specific easing implementation
    class.
    """
    may_overshoot: bool
    """
    Boolean flag to indicate if the easing function can
    generate values out of the bounds or not.
    """

    def __init_subclass__(
        cls,
        **kwargs
    ):
        """
        Init the subclass and register it.
        """
        super().__init_subclass__(**kwargs)

        if hasattr(cls, '_name'):
            EasingFunction._registry[cls.get_name()] = cls

    @classmethod
    def get(
        cls,
        name: EasingFunctionName
    ) -> type['EasingFunction']:
        """
        Get the specific class associated to the given
        easing animated function `name`.
        """
        name = EasingFunctionName.to_enum(name)

        return cls._registry[name.value]

    @classmethod
    def get_name(
        cls
    ) -> str:
        """
        Get the string name of the specific easing
        implementation class.
        """
        return cls._name

    @abstractmethod
    def _ease(
        self,
        t_normalized: float,
        **kwargs
    ) -> float:
        """
        *For internal use only*

        The proper calculation of the easing, that must be
        implemented for each specific class.

        The ease calculation will provide the normalized
        value (in `[0.0, 1.0]` range) for the given 
        `t_normalized` time (in `[0.0, 1.0]` range also),
        that is the value (but normalized) associated to
        that time moment.
        """
        pass

    def ease(
        self,
        t_normalized: float,
        do_clamp: bool = True,
        **kwargs
    ) -> float:
        """
        Apply the easing to the `t_normalized` value provided,
        that must be a value in the `[0.0, 1.0]` range, to
        obtain the normalized value associated to it.

        The value will be clamp if the `do_clamp` parameter is
        `True` and it's needed.

        Some easing functions could return values that are out
        of the limits (`value < 0.0` or `value > 1.0`), which
        is called 'overshoot'. The clamp will transform those
        values out of bounds into the limits, turning `1.1`
        into `1.0` or `-0.1` into `0.0`.
        """
        ParameterValidator.validate_mandatory_number_between(
            name = 't_normalized',
            value = t_normalized,
            lower_limit = 0.0,
            upper_limit = 1.0,
            do_include_lower_limit = True,
            do_include_upper_limit = True
        )
        # Value is already clamp if no exception raise

        value_eased = self._ease(
            t_normalized = t_normalized,
            **kwargs
        )

        return (
            clamp_value_normalized(value_eased)
            if do_clamp else
            value_eased
        )

    def __call__(
        self,
        t_normalized: float,
        do_clamp: bool = True,
        **kwargs
    ) -> float:
        """
        Special method to allow us doing `ease_function(alpha)`
        instead of `ease_function.ease(alpha)`.

        Apply the easing to the `t_normalized` value provided,
        that must be a value in the `[0.0, 1.0]` range, to
        obtain the normalized value associated to it.

        The value will be clamp if the `do_clamp` parameter is
        `True` and it's needed.

        Some easing functions could return values that are out
        of the limits (`value < 0.0` or `value > 1.0`), which
        is called 'overshoot'. The clamp will transform those
        values out of bounds into the limits, turning `1.1`
        into `1.0` or `-0.1` into `0.0`.
        """
        return self.ease(
            t_normalized = t_normalized,
            do_clamp = do_clamp,
            **kwargs
        )

    @requires_dependency('matplotlib', 'yta_math', 'matplotlib')
    def plot(
        self,
        number_of_samples: int = 100
    ):
        """
        Plot a graphic representing `number_of_samples`
        values, from `0.0` to `1.0`, to see how the easing
        function works.
        """
        import matplotlib.pyplot as plt
        
        # Limit and draw axis
        plt.xlim(0, 1)
        plt.ylim(0, 1)
        plt.axhline(0, color = 'black', linewidth = 1)
        plt.axvline(0, color = 'black', linewidth = 1)

        plt.grid(True)

        # We will draw 'n'' values between 0.0 and 1.1
        x_vals = np.linspace(0.0, 1.0, number_of_samples).tolist()
        y_vals = [
            self.get_n_value(x)
            for x in x_vals
        ]
        plt.scatter(x_vals, y_vals, color = 'white', edgecolors = 'black', s = 100)

        # TODO: I think we don't need to draw anything
        # else. Remove this below if it is ok.
        # # Draw points between nodes
        # for pair_of_node in self.pairs_of_nodes:
        #     positions = pair_of_node.get_n_xy_values_to_plot(100)
        #     t_xs, t_ys = zip(*positions)
        #     xs += t_xs
        #     ys += t_ys
       
        # plt.scatter(xs, ys, color = 'black', s = 1)
        
        plt.title('')
        plt.xlabel('X axis')
        plt.ylabel('Y axis')
        
        plt.show()



from yta_math_easings.abstract import EasingFunction
from yta_math_easings.enums import EasingFunctionName


class SmoothererStepEasing(EasingFunction):
    """
    A smootherer step easing function.

    Implementation of the 3rd order SmoothStep sigmoid function.
    The 1st, 2nd and 3rd derivatives (speed, acceleration and jerk)
    are zero at the endpoints.
    https://en.wikipedia.org/wiki/Smoothstep

    The formula:
    ```
    (
        35 * n**4 - 84 * n**5 + 70 * n**6 - 20 * n**7
        if 0 < n < 1 else
        1
        if n >= 1 else
        0
    )
    ```
    """
    _name: str = EasingFunctionName.SMOOTHERER_STEP.value

    def ease(
        self,
        t_normalized: float,
        do_clamp: bool = True
    ) -> float:
        return super().ease(
            t_normalized = t_normalized,
            do_clamp = do_clamp
        )

    def _ease(
        self,
        t_normalized: float
    ) -> float:
        return (
            35 * t_normalized**4 - 84 * t_normalized**5 + 70 * t_normalized**6 - 20 * t_normalized**7
            if 0 < t_normalized < 1 else
            1
            if t_normalized >= 1 else
            0
        )


from yta_math_easings.abstract import EasingFunction
from yta_math_easings.enums import EasingFunctionName


class LinearEasing(EasingFunction):
    """
    A linear easing function.

    The formula:
    ```
    `t_normalized`
    ```
    """
    _name: str = EasingFunctionName.LINEAR.value

    def ease(
        self,
        t_normalized: float,
        do_clamp: bool = True
    ) -> float:
        return super().ease(
            t_normalized = t_normalized,
            do_clamp = do_clamp
        )

    def _ease(
        self,
        t_normalized: float
    ) -> float:
        return t_normalized
    



"""
Module to include the functionality related
to easings.

Thanks for the inspiration:
- https://github.com/semitable/easing-functions/blob/master/easing_functions
"""
from yta_math_easings.linear import LinearEasing
from yta_math_easings.slow_into import SlowIntoEasing
from yta_math_easings.smooth import SmoothEasing
from yta_math_easings.smooth_step import SmoothStepEasing
from yta_math_easings.smoother_step import SmootherStepEasing
from yta_math_easings.smootherer_step import SmoothererStepEasing

# TODO: Add more from 'yta_math.rate_functions'

"""
Here you have a list including the 'overshoot' of
the different easing functions, including also the
[X] sign for those that have been already included
in the library.

[X] linear - no
[X] smooth — no
[X] smooth_step — no
[X] smoother_step — no
[X] smootherer_step — no
[ ] rush_into — no
[ ] rush_from — no
[ ] double_smooth — no
[ ] there_and_back — no
[ ] there_and_back_with_pause — no
[ ] wiggle — sí
[ ] ease_in_cubic — no
[ ] ease_out_cubic — no
[ ] ease_in_out_cubic — no
[ ] ease_in_quad — no
[ ] ease_out_quad — no
[ ] ease_in_out_quad — no
[ ] ease_in_quart — no
[ ] ease_out_quart — no
[ ] ease_in_out_quart — no
[ ] ease_in_quint — no
[ ] ease_out_quint — no
[ ] ease_in_out_quint — no
[ ] ease_in_expo — no
[ ] ease_out_expo — no
[ ] ease_in_out_expo — no
[ ] ease_in_sine — no
[ ] ease_out_sine — no
[ ] ease_in_out_sine — no
[ ] ease_in_circ — no
[ ] ease_out_circ — no
[ ] ease_in_out_circ — no
[ ] ease_in_back — sí
[ ] ease_out_back — sí
[ ] ease_in_out_back — sí
[ ] ease_in_elastic — sí
[ ] ease_out_elastic — sí
[ ] ease_in_out_elastic — sí
[ ] ease_in_bounce — no
[ ] ease_out_bounce — no
[ ] ease_in_out_bounce — no
[ ] squish_rate_func — sí (depende de func)
[ ] lingering — no
[ ] exponential_decay — no
"""


__all__ = [
    'LinearEasing',
    'SlowIntoEasing',
    'SmoothEasing',
    'SmoothStepEasing',
    'SmootherStepEasing',
    'SmoothererStepEasing'
]





from yta_math_easings.animated.abstract import EasingAnimatedFunction
from yta_math_easings.smoother_step import SmootherStepEasing


class SmootherStepAnimatedEasing(EasingAnimatedFunction):
    """
    A smoother step function but animated, including a
    range of values and a duration.

    The formula:
    ```
    (
        0
        if t_normalized <= 0 else
        (
            6 * t_normalized**5 - 15 * t_normalized**4 + 10 * t_normalized**3
            if t_normalized < 1 else
            1
        )
    )
    """
    may_overshoot: bool = False
    _easing_cls: type['EasingFunction'] = SmootherStepEasing

    def ease(
        self,
        t_normalized: float,
        do_clamp: bool = True
    ) -> float:
        return super().ease(
            t_normalized = t_normalized,
            do_clamp = do_clamp
        )


from yta_math_easings.progress_mapping.abstract import ProgressMapping


class ReverseProgressMapping(ProgressMapping):
    """
    The reverse progress mapping, that will return
    the opositve value than the one provided.

    The formula:
    ```
    1.0 - u
    ```
    """

    def map(
        self,
        u: float
    ) -> float:
        return super().map(
            u = u
        )

    def _map(
        self,
        u: float
    ) -> float:
        return 1.0 - u